#include "schedos-kern.h"
#include "x86struct.h"
#include "x86.h"

/*****************************************************************************
 * schedos-kern
 *
 *   This is the schedos's kernel.
 *   It sets up process control blocks for the 4 applications, then runs
 *   them in some schedule.
 *
 *****************************************************************************/

// The bootloader loads 4 processes, starting at PROCESS_START, allocating
// 1 MB to each process.
// Each process's stack grows down from the top of its memory space.

#define NPROCESSES	4
#define PROCESS_START	0x200000
#define PROCESS_SIZE	0x100000

// +--------------+------------------------+----------------------+-----------/
// | Base Memory  | Kernel          Kernel | App 0          App 0 | App 1
// |              | Code + Data      Stack | Code + Data    Stack | Code ...
// +--------------+------------------------+----------------------+-----------/
// 0x0        0x100000                 0x200000               0x300000
//
// When the boot loader exits, each application's starting instruction pointer
// is stored at the very top of its stack.
//
// There is also a shared 'cursorpos' variable, located at 0x190000 in the
// kernel's data area.


// A process control block for each process.
static pcb_t pcbs[NPROCESSES];

// The currently running process ID.
int current_pid;

// The preferred scheduling algorithm.
int scheduling_algorithm;


/*****************************************************************************
 * start
 *
 *   Initialize the hardware and process control blocks, then run
 *   the first process.
 *
 *****************************************************************************/

void
start(void)
{
	int i;

	// Set up hardware (schedos-x86.c)
	segment_descriptors_init();
	interrupt_controller_init(0);
	console_clear();

	// Set up process control blocks (the pcbs[] array)
	for (i = 0; i < NPROCESSES; i++) {
		pcb_t *pcb = &pcbs[i];
		uint32_t stack_ptr = PROCESS_START + (i + 1) * PROCESS_SIZE;

		// Initialize the process control block
		pcb_init(pcb);
		
		// Set ESP and (using value set by bootloader) EIP
		pcb->pcb_registers.reg_esp = stack_ptr;
		pcb->pcb_registers.reg_eip = *((uint32_t *) (stack_ptr - 4));

		// Mark the process as runnable!
		pcb->pcb_runnable = 1;
	}

	// Initialize the cursor-position shared variable to point to the
	// console's first character (the upper left).
	cursorpos = (uint16_t *) 0xB8000;

	// Initialize the scheduling algorithm.
	scheduling_algorithm = 0;

	// Switch to the first process.  The context_switch() function
	// context switches to a particular PCB, which causes that process
	// to start running.
	current_pid = 0;
	context_switch(&pcbs[0]);

	// Should never get here!
	while (1)
		/* do nothing */;
}



/*****************************************************************************
 * trap
 *
 *   This is the weensy interrupt and system call handler.
 *   The current handler handles 4 different system calls (two of which
 *   do nothing), plus the clock interrupt.
 *
 *   Note that we will never receive clock interrupts while in the kernel.
 *
 *****************************************************************************/

void
trap(registers_t *reg)
{
	// Save the current process's register state into the PCB.
	pcbs[current_pid].pcb_registers = *reg;

	switch (reg->reg_trapno) {

	    case TRAP_SYS_YIELD:
		// The 'sys_yield' system call asks the kernel to schedule
		// the next process.
		schedule();

	    case TRAP_SYS_EXIT:
		// 'sys_exit' exits the current process: it is marked as
		// non-runnable.
		// The application stored its exit status in the %eax register
		// before calling the system call.  The %eax register has
		// changed by now, but we can read the application's value
		// out of the 'reg' argument.
		// (This shows you how to transfer arguments to system calls!)
		pcbs[current_pid].pcb_runnable = 0;
		pcbs[current_pid].pcb_exit_status = reg->reg_eax;
		schedule();

	    case TRAP_SYS_USER1:
		// 'sys_user*' are provided for your convenience, in case you
		// want to add a system call.
		/* Your code here (if you want). */
		context_switch(&pcbs[current_pid]);

	    case TRAP_SYS_USER2:
		/* Your code here (if you want). */
		context_switch(&pcbs[current_pid]);

	    case TRAP_CLOCKINTERRUPT:
		// A clock interrupt occurred (so an application exhausted its
		// time quantum).
		// Switch to the next runnable process.
		schedule();

	    default:
		while (1)
			/* do nothing */;

	}
}



/*****************************************************************************
 * schedule
 *
 *   This is the weensy process scheduler.
 *   It picks a runnable process, then context-switches to that process.
 *   If there are no runnable processes, it spins forever.
 *
 *   This function implements multiple scheduling algorithms, depending on
 *   the value of 'scheduling_algorithm'.  We've provided one; in the problem
 *   set you will provide at least one more.
 *
 *****************************************************************************/

void
schedule(void)
{
	if (scheduling_algorithm == 0)
		while (1) {
			current_pid = (current_pid + 1) % NPROCESSES;

			// Run the selected process!  
			// But ignore non-runnable processes.
			// Note that the 'context_switch' function
			// does not return.
			if (pcbs[current_pid].pcb_runnable)
				context_switch(&pcbs[current_pid]);
		}

	// If we get here, we are running an unknown scheduling algorithm.
	while (1)
		/* do nothing */;
}
